package pku.edu.hocra.graph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import pku.edu.hocra.graph.ds.IncrementalIndexing;
import pku.edu.hocra.graph.ds.IntPair;
import soot.Local;
import soot.SootMethod;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.jimple.AssignStmt;
import soot.jimple.DefinitionStmt;
import soot.jimple.InterfaceInvokeExpr;
import soot.jimple.InvokeExpr;
import soot.jimple.ParameterRef;
import soot.jimple.ReturnStmt;
import soot.jimple.Stmt;
import soot.jimple.VirtualInvokeExpr;
import soot.jimple.internal.JInvokeStmt;



public class ProceduralGraph {
	
	private IncrementalIndexing<Local> localIndex = new IncrementalIndexing<Local>();
	private int n;
	
	private class EdgeSet {
		private HashSet<Integer> edges = new HashSet<Integer>();
		void addEdge(int u, int v) {
			edges.add(u * n + v);
		}
		List<IntPair> getEdges() {
			ArrayList<IntPair> e = new ArrayList<IntPair>();
			for (Integer i : edges) {
				e.add(new IntPair(i/n, i%n));
			}
			return e;
		}
	}
	
	EdgeSet edgeSet = new EdgeSet();
	public Map<Integer, Integer> entryPrmt2Local = new HashMap<Integer, Integer>();
	public Set<Integer> exitLocals = new HashSet<Integer>();
	public Map<Integer, HashMap<Integer, Integer>> callStmtPrmt2Local = new HashMap<Integer, HashMap<Integer, Integer>>();
	public Map<Integer, Integer> retStmt2Local = new HashMap<Integer, Integer>();
	public Set<IntPair> callbackNodePairSet = new HashSet<IntPair>();
	
	public int getNodeNumber() {
		return n;
	}
	
	public List<IntPair> getEdges() {
		return edgeSet.getEdges(); 
	}
	
	public ProceduralGraph(SootMethod m) {
				
		for (Local l : m.getActiveBody().getLocals()) {
			localIndex.append(l);
		}
		n = localIndex.size();
		
		// Intraprocedural edges
		for (Unit u : m.getActiveBody().getUnits()) {
			if (u instanceof AssignStmt) {
				AssignStmt stmt = (AssignStmt) u;
				if (!(stmt.getRightOp() instanceof InvokeExpr)) {
					Value leftValue = stmt.getLeftOp();
					Value rightValue = stmt.getRightOp();
					for (Integer li : extractLocals(leftValue))
						for (Integer ri : extractLocals(rightValue)) {
							//System.out.println(ri + "->" + li);
							edgeSet.addEdge(ri, li);
						}
				}
			}
		}
		
		// EntrySites
		for (Unit u : m.getActiveBody().getUnits()) {
			if (u instanceof DefinitionStmt) {
				DefinitionStmt defu = (DefinitionStmt) u;
				if (defu.getLeftOp() instanceof Local && defu.getRightOp() instanceof ParameterRef) {
					Local leftOp = (Local) defu.getLeftOp();
					ParameterRef rightOp = (ParameterRef) defu.getRightOp();
					int parameterIndex = rightOp.getIndex();
					//System.out.println("EP(" + parameterIndex + ") " + localIndex.getIndex(leftOp));
					entryPrmt2Local.put(parameterIndex, localIndex.getIndex(leftOp));
				}
			}
		}
		
		// ExitSites
		for (Unit u : m.getActiveBody().getUnits()) {
			if (u instanceof ReturnStmt) {
				ReturnStmt retStmt = (ReturnStmt) u;
				if (retStmt.getOp() instanceof Local) {
					Local local = (Local) retStmt.getOp();
					//System.out.println("Ex " + localIndex.getIndex(local));
					exitLocals.add(localIndex.getIndex(local));
				}
			}
		}
		
		// CallSites
		int i = 0;
		for (Unit u : m.getActiveBody().getUnits()) {
			InvokeExpr invokeExpr = null;
			if (u instanceof JInvokeStmt) {
				Stmt stmt = (Stmt) u;
				invokeExpr = stmt.getInvokeExpr();
			}
			else if (u instanceof DefinitionStmt) {
				DefinitionStmt defStmt = (DefinitionStmt) u;
				if (defStmt.getRightOp() instanceof InvokeExpr) {
					invokeExpr = (InvokeExpr) defStmt.getRightOp(); 
				}
			}
			if (invokeExpr != null) {
				final List<Value> args = invokeExpr.getArgs();
				int j = 0;
				for (Value value : args) {
					if (value instanceof Local) {
						Local local = (Local) value;
						//System.out.println("I(" + i + ") CP(" + j + ") " + localIndex.getIndex(local));
						//callStmtPrmt2Local.put(new IntPair(i, j), localIndex.getIndex(local));
						if (!callStmtPrmt2Local.containsKey(i)) {
							callStmtPrmt2Local.put(i, new HashMap<Integer, Integer>());
						}
						HashMap<Integer, Integer> t = callStmtPrmt2Local.get(i);
						t.put(j, localIndex.getIndex(local));
					}
					j ++;
				}
			}
			i ++;
		}
		
		// ReturnSites
		i = 0;
		for (Unit u : m.getActiveBody().getUnits()) {
			if (u instanceof DefinitionStmt) {
				DefinitionStmt defu = (DefinitionStmt) u;
				if (defu.getRightOp() instanceof InvokeExpr && defu.getLeftOp() instanceof Local) {
					Local leftOp = (Local) defu.getLeftOp();
					//System.out.println("I(" + i + ") R " + localIndex.getIndex(leftOp));
					retStmt2Local.put(i, localIndex.getIndex(leftOp));
				}
			}
			i ++;
		}
		
		// Callbacks
		for (Entry<Integer, HashMap<Integer, Integer>> et : callStmtPrmt2Local.entrySet()) {
			if (retStmt2Local.containsKey(et.getKey())) {
				Integer rt = retStmt2Local.get(et.getKey());
				for (Integer lc : et.getValue().values()) {
					callbackNodePairSet.add(new IntPair(lc, rt));
				}
			}
		}
		
	}
	
	private Collection<Integer> extractLocals(Value expr) {
		Collection<Integer> result = new HashSet<Integer>();
		if (expr instanceof Local) {
			result.add(localIndex.getIndex((Local) expr));
		}
		else {
			for (ValueBox vb : expr.getUseBoxes()) {
				Value v = vb.getValue();
				if (v instanceof Local)
					result.add(localIndex.getIndex((Local) v));
			}
		}
		return result;		
	}
	
	/*
	private void operateValue(Set<String> classSet, Value value) {
		if (!classSet.contains(value.getClass().toString())) {
			System.out.print(classSet.size() + " ");
			System.out.println(value.getClass().toString());
			System.out.println("Example: " + value);
			System.out.print("         ");
			for (ValueBox v : value.getUseBoxes())
				System.out.print(v.getValue() + "[" + v + "] ");
			System.out.println();
			classSet.add(value.getClass().toString());
		}
	}
	*/
	
}
